/* branching.c
 *
 * Perform the conversion from a DTWA into a Bottom-Up Tree Automaton.
 *
 * Copyright (C) 2009 ENS Cachan
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA.
 */
#include <assert.h>
#include <stdbool.h>
#include <stdlib.h>

#include <glib.h>
#include <libxml/hash.h>

#include "bitset.h"
#include "branching.h"
#include "buta.h"
#include "dtwa.h"

// search for transitions that go in a given direction
static GSList *
get_transitions_towards(const dtwa *wa, state p, type t, direction d)
{
  symbol a;
  destination *dest;
  GSList *l = NULL;
  
  for (a = 0; a < wa->nsyms; a++)
    if (exists_transition_from(wa, p, a, t))
      {
        dest = wa->trans[a][p][t];
        if (dest->d == d)
          l = g_slist_prepend(l, GUINT_TO_POINTER(dest->s));
      }
  return l;
}

// Extended states
typedef struct
{
  type t;
  bitset *pairs;
  guint size;
  guint id;
  guint hash;
} extstate;

static extstate *
new_extstate(unsigned int size, type t)
{
  extstate *s = malloc (sizeof (*s));
  unsigned int i;
  if (s == NULL)
    {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
    }
  s->t = t;

  s->pairs = calloc(size, sizeof(*s->pairs));
  if (s->pairs == NULL)
    {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
    }
  for (i = 0; i < size; i++)
    s->pairs[i] = NULL;

  s->size = size;
  s->id = 0;
  s->hash = t;
  return s;
}

static void
add_pair_extstate(state p, state q, extstate *s)
{
  assert(s != NULL);
  assert(s->pairs != NULL);

  if (s->pairs[p] == NULL)
    s->pairs[p] = new_bitset(s->size);

  set_bitset(s->pairs[p], q);
  s->hash = 31 * s->hash + 13 * p + q;
}

static inline bool
exists_pair_extstate(state p, state q, extstate *s)
{
  assert(s != NULL);
  assert(s->pairs != NULL);
  
  return s->pairs[p] != NULL && test_bitset(s->pairs[p], q);
}

static guint
hash_extstate (const extstate *s)
{
  assert(s != NULL);
  return s->hash;
}

static gboolean
equal_extstates (const extstate *s1, const extstate *s2)
{
  gboolean ret = TRUE;
  unsigned int i;
  assert(s1 != NULL);
  assert(s2 != NULL);
  assert(s1->size == s1->size);

  if (s1->t != s2->t)
    return FALSE;

  for (i = 1; ret && i < s1->size; i++)
    ret = ret && equal_bitsets(s1->pairs[i], s2->pairs[i], s1->size);

  return ret;
}

static void
free_extstate (extstate *s)
{
  unsigned int i;
  assert(s != NULL);
  for (i = 0; i < s->size; i++)
    if (s->pairs[i] != NULL)
      free(s->pairs[i]);
  free (s->pairs);
  free (s);
  s = NULL;
}

typedef struct
{
  char **vars;
  char **states;
} names_t;

static void
record_extstate (extstate *e, gpointer value, names_t *vars)
{
  GString *ret = g_string_new("v");
  unsigned int i, j;

  g_string_append_printf(ret, "%u_", e->id);
  g_string_append(ret, get_type_name(e->t));
  for (i = 0; i < e->size; i++)
    if (e->pairs[i] != NULL)
      for (j = 0; j < e->size; j++)
        if (i != j && test_bitset(e->pairs[i], j))
          g_string_append_printf(ret,
                                 "_%s-%s",
                                 get_state_name(i, vars->states),
                                 get_state_name(j, vars->states));
  vars->vars[e->id] = g_string_free(ret, FALSE);
}

// Roy-Warshall transitive closure on an extended state
static void
transitive_closure_extstate(extstate *e)
{
  unsigned int i, j, k;
  assert(e != NULL);

  for (i = 0; i < e->size; i++)
    add_pair_extstate(i, i, e);
  
  for (k = 0; k < e->size; k++)
    for (i = 0; i < e->size; i++)
      for (j = 0; j < e->size; j++)
        if (exists_pair_extstate(i, k, e)
            && exists_pair_extstate(k, j, e))
          add_pair_extstate(i, j, e);
}

// Real things start here...

// Create an extended state with pairs (q, p) such that
//    q, a, t --> p, STAY
static inline extstate*
create_stay_extstate(dtwa* wa, symbol a, type t)
{
  destination *dest;
  state q;
  extstate *e;
  e = new_extstate(wa->nstates, t);

  // initial pairs
  for (q = 1; q < wa->nstates; q++)
    if (exists_transition_from(wa, q, a, t))
      {
        dest = wa->trans[a][q][t];
        if (dest->d == STAY)
          add_pair_extstate(q, dest->s, e);
      }
  return e;
}

// Complete an extended state `e' with pairs (q, s) such that
//   1.  q, a, t  --> p, d
//   2.  (p, r) is a pair of `from'
//   3.  r, _, t' --> s, UP
//      (with `_' a wildcard and t' the type of `from')
static void
complete_extstate(extstate *e, dtwa *wa, symbol a, type t, direction d,
                  extstate *from)
{
  state p;
  assert(from != NULL);
  assert(e != NULL);
  assert(wa != NULL);
  assert(d == DOWN0 || d == DOWN1);

  for (p = 0; p < from->size; p++)
    if (from->pairs[p] != NULL)
      {
        destination **dests = wa->reach[p];
        unsigned int i;

        for (i = 1; i <= (int) dests[0]; i++)
          if (dests[i]->d == reverse_direction(d))
            // if we find a transition to `p' with appropriate symbol and type
            if (exists_transition_to(wa, dests[i]->s, a, t, p, d))
              {
                state r;
                for (r = 1; r < from->size; r++)
                  if (test_bitset(from->pairs[p], r))
                    {
                      // find transitions from `r'
                      GSList *c, *l;
                      c = l = get_transitions_towards(wa, r, from->t, UP);
                      while (c != NULL)
                        {
                          add_pair_extstate(dests[i]->s,
                                            GPOINTER_TO_UINT(c->data),
                                            e);
                          c = c->next;
                        }
                      g_slist_free(l);
                    }
              }
      }
}

static void
sym_copy (void * payload, void * data, xmlChar * name)
{
  char **syms = (char**) data;
  syms[(int)payload - 1] = (char*) name;
}

buta *
dtwa2buta (dtwa* wa, bool names, bool notrans)
{
  buta *b;
  GHashTable *states;
  GSList *processed[2] = { NULL, NULL };
  GSList *unprocessed = NULL;
  symbol a;
  type t;
  extstate *e;
  unsigned int count = 1;
  assert(wa != NULL);
  
  // initialization
  b = new_buta();
  states = g_hash_table_new_full((GHashFunc) hash_extstate,
                                 (GEqualFunc) equal_extstates,
                                 (GDestroyNotify) free_extstate,
                                 NULL);

  // getting the leaves
  for (a = 0; a < wa->nsyms; a++)
    for (t = LROOT; t < NTYPES; t++)
      if (t == LROOT || t == L0 || t == L1)
        {
          e = create_stay_extstate(wa, a, t);
          // transitive closure
          transitive_closure_extstate(e);
          // is it new?
          if (!g_hash_table_lookup(states, e))
            {
              g_hash_table_insert(states, e, GUINT_TO_POINTER(1));
              e->id = count++;

              // add to the lists
              if (t == L0 || t == L1)
                unprocessed = g_slist_prepend(unprocessed, e);

              // add transition to a constant
              if (!notrans)
                add_rule(b, a, 0, 0, e->id);
                
              // set as initial
              if (t == LROOT && exists_pair_extstate(INIT, FINAL, e))
                set_initial(b, e->id);
            }
          else
            free_extstate(e);
        }
  
  // main loop
  while (unprocessed != NULL)
    {
      extstate *u1 = (extstate *) unprocessed->data;
      GSList *current;
      gboolean c0 = u1->t % 2;
      unprocessed = g_slist_remove(unprocessed, u1);
      processed[c0] = g_slist_prepend(processed[c0], u1);

      current = processed[(c0+1)%2];
      while (current != NULL)
        {
          extstate *u2 = (extstate *) current->data;

          for (a = 0; a < wa->nsyms; a++)
            for (t = ROOT; t < L0; t++)
              if (t == ROOT || t == C0 || t == C1)
                {
                  guint i;
                  // get a fresh extended state with stay pairs
                  e = create_stay_extstate(wa, a, t);
                  
                  // complete the extended state with data from u1
                  complete_extstate(e, wa, a, t, c0 + DOWN0, u1);

                  // same with data from u2
                  complete_extstate(e, wa, a, t, c0 + DOWN0, u2);

                  // transitive closure
                  transitive_closure_extstate(e);
                  
                  // is it new?
                  if (!(i = GPOINTER_TO_UINT(g_hash_table_lookup(states, e))))
                    {
                      e->id = count++;
                      i = e->id;
                      g_hash_table_insert(states, e, GUINT_TO_POINTER(e->id));

                      // add to the list
                      if (t == C0 || t == C1)
                        unprocessed = g_slist_prepend(unprocessed, e);

                      // add as initial
                      if (t == ROOT && exists_pair_extstate(INIT, FINAL, e))
                        set_initial(b, e->id);
                    }
                  else
                    free_extstate(e);

                  // add transition
                  if (!notrans)
                    {
                      if (c0)
                        {
                          // add the transition a(u2, u1) -> e
                          add_rule(b, a, u2->id, u1->id, i);
                        }
                      else
                        {
                          // add the transition a(u1, u2) -> e
                          add_rule(b, a, u1->id, u2->id, i);
                        }
                    }
                }
          
          current = current->next;
        }
    }  

  b->nvars = count - 1;
  b->nsyms = wa->nsyms;

  // record symbol names
  if (wa->symbols != NULL)
    {
      xmlHashTablePtr hash = (xmlHashTablePtr) wa->symbols;
      b->symbols = calloc(b->nsyms, sizeof(*b->symbols));
      if (b->symbols == NULL)
        {
          fprintf(stderr, "Out of memory.\n");
          exit(1);
        }
      xmlHashScan(hash, sym_copy, b->symbols);
    }

  // record var names
  if (names)
    {
      b->vars = calloc(b->nvars+1, sizeof(*b->vars));
      if (b->vars == NULL)
        {
          fprintf(stderr, "Out of memory.\n");
          exit(1);
        }
      else
        {
          names_t n = { b->vars, wa->states };
          g_hash_table_foreach(states, (GHFunc) record_extstate, &n);
        }
    }
  
  // cleaning up
  g_slist_free(unprocessed);
  g_slist_free(processed[0]);
  g_slist_free(processed[1]);
  g_hash_table_destroy(states);

  return b;
}
/*
// Pairs of states
typedef struct
{
  state v1;
  state v2;
} pair;

static inline pair*
new_pair (state v1, state v2)
{
  pair *p = malloc(sizeof (*p));
  if (p == NULL)
    {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
    }
  p->v1 = v1;
  p->v2 = v2;
  return p;
}

static guint
hash_pair (const pair *p)
{
  return p->v1 * 61 + p->v2;
}

static gboolean
equal_pairs (const pair*p1, const pair *p2)
{
  return p1->v1 == p2->v1 && p1->v2 == p2->v2;
}

static void
free_pair (pair *p)
{
  free (p);
  p = NULL;
}

typedef struct
{
  state GHashTable *table;
  void *incl;
} run;

static void
check_table_inclusion (gpointer key, gpointer value, gpointer user_data)
{
  run *r;
  assert(user_data != NULL);
  r = (run *)user_data;
  if (r->incl)
    if (!g_hash_table_lookup(r->table, key))
      r->incl = NULL;
}

static gboolean
equal_extstates (const extstate *s1, const extstate *s2)
{
  assert(s1 != NULL);
  assert(s2 != NULL);
  
  if (s1->t != s2->t)
    return FALSE;

  if (g_hash_table_size (s1->pairs) != g_hash_table_size (s2->pairs))
    return FALSE;
  
  r.table = s2->pairs;
  r.incl  = (void*)1;
  g_hash_table_foreach(s1->pairs, check_table_inclusion, &r);
  if (r.incl == NULL)
    return FALSE;

  r.table = s1->pairs;
  r.incl  = (void*)1;
  g_hash_table_foreach(s2->pairs, check_table_inclusion, &r);
  
  return r.incl != NULL;
}

static void
free_extstate (extstate *s)
{
  assert(s != NULL);
  g_hash_table_destroy(s->pairs);
  free (s);
  s = NULL;
}
*/
