/* buta.c
 *
 * Implementation of binary bottom-up tree automata.
 *
 * Copyright (C) 2009 ENS Cachan
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA.
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <glib.h>

#include "buta.h"

// Entries in our hash tables
typedef struct
{
  unsigned int sym;
  var v1;
  var v2;
  guint hash; // cached hash
} entry;

static entry *
new_entry (unsigned int sym, var v1, var v2)
{
  entry *ret = malloc(sizeof(*ret));
  if (ret == NULL)
    {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
    }
  ret->sym = sym;
  ret->v1  = v1;
  ret->v2  = v2;
  ret->hash = ((sym * 31) + v1) * 31 + v2;

  return ret;
}

static guint
hash_entry (const entry* e)
{
  assert(e != NULL);
  return e->hash;
}

static gboolean
equal_entries (const entry *a, const entry *b)
{
  assert(a != NULL);
  assert(b != NULL);

  return (a->sym == b->sym)
    && (a->v1 == b->v1)
    && (a->v2 == b->v2);
}

static void
free_entry(entry *e)
{
  free(e);
  e = NULL;
}

// Creating binary BUTA
static guint
hash_var (const void* e)
{
  var v;
  assert(e != NULL);
  v = GPOINTER_TO_UINT(e);
  return v % 127;
}

static gboolean
equal_vars (const void *a, const void *b)
{
  assert(a != NULL);
  assert(b != NULL);
  return GPOINTER_TO_UINT(a) == GPOINTER_TO_UINT(b);
}

buta*
new_buta (void)
{
  buta *ret = malloc(sizeof(*ret));
  if (ret == NULL)
    {
      fprintf(stderr, "Out of memory.\n");
      exit(1);
    }
  ret->nvars   = 0;
  ret->nsyms   = 0;
  ret->rules   = NULL;//calloc(1, sizeof(*ret->rules));
  ret->initial = g_hash_table_new((GHashFunc) hash_var,
                                  (GEqualFunc) equal_vars);
  ret->symbols = NULL;
  ret->vars    = NULL;
  return ret;
}

void
add_rule (buta* b, unsigned int a, var v1, var v2, var v)
{
  entry *e;
  assert(b != NULL);

  // grow the number of variables if needed
  if (v >= b->nvars)
    {
      unsigned int i;
      b->rules = realloc(b->rules, (v+1)*sizeof(*b->rules));
      if (b->rules == NULL)
        {
          fprintf(stderr, "Out of memory.\n");
          exit(1);
        }

      for (i = b->nvars + 1; i <= v; i++)
        b->rules[i] = g_hash_table_new_full ((GHashFunc) hash_entry, 
                                             (GEqualFunc) equal_entries,
                                             (GDestroyNotify) free_entry,
                                             NULL);
      b->nvars = v;
      fprintf(stderr, "%u\n", b->nvars);
    }

  // remember the number of symbols
  if (a >= b->nsyms)
    b->nsyms = a + 1;

  // create a new entry
  e = new_entry(a, v1, v2);
  if (g_hash_table_lookup(b->rules[v], e))
    free_entry(e);
  else
    g_hash_table_insert(b->rules[v], e, GUINT_TO_POINTER(1));
}

inline void
set_initial (buta *b, var v)
{
  assert(b != NULL);
  g_hash_table_insert(b->initial,
                      GUINT_TO_POINTER(v),
                      GUINT_TO_POINTER(1));
}

void
free_buta (buta *b)
{
  unsigned int i;
  assert(b != NULL);
  if (b->rules != NULL) 
    for (i = 1; i <= b->nvars; i++)
      g_hash_table_destroy(b->rules[i]);
  free(b->rules);
  g_hash_table_destroy(b->initial);
  if (b->vars != NULL)
    free(b->vars);
  if (b->symbols != NULL)
    free(b->symbols);
  free(b);
  b = NULL;
}


// RELAX NG output
typedef struct
{
  buta *fst;
  FILE *snd;
  void *thrd;
} triple;

static inline void
print_var (buta *b, var v, FILE *out)
{
  if (b->vars != NULL)
    fprintf(out, "%s", b->vars[v]);
  else
    fprintf(out, "v%u", v);
}

static inline void
print_symbol (buta *b, unsigned int s, FILE *out)
{
  assert(s < b->nsyms);
  if (b->symbols != NULL)
    fprintf(out, "%s", ((char **)b->symbols)[s]);
  else
    fprintf(out, "s%u", s);
}

static void
relax_ng_print_var (gpointer key, gpointer value, triple *p)
{
  guint v;
  assert(p != NULL);
  if (p->thrd)
    {
      fprintf(p->snd, "    ");
      p->thrd = NULL;
    }
  else
    fprintf(p->snd, "  | ");

  v = GPOINTER_TO_UINT(key);
  print_var (p->fst, v, p->snd);
  fprintf(p->snd, "\n");
}

static void
relax_ng_print_rhs(gpointer key, gpointer value, triple *p)
{
  entry *e;
  assert(p != NULL);
  if (p->thrd)
    {
      fprintf(p->snd, "    ");
      p->thrd = NULL;
    }
  else
    fprintf(p->snd, "  | ");

  e = (entry *) key;
  fprintf(p->snd, "element ");
  print_symbol(p->fst, e->sym, p->snd);
  fprintf(p->snd, " { ");
  if (e->v1 == 0 && e->v2 == 0)
    fprintf(p->snd, "empty");
  else
    {
      print_var(p->fst, e->v1, p->snd);
      fprintf(p->snd, ", ");
      print_var(p->fst, e->v2, p->snd);
    }
  fprintf(p->snd, " }\n");
}

void
relax_ng_print_buta (buta *b, FILE *out)
{
  triple t = { b, out, (void *)1 };
  unsigned int i;

  fprintf(out, "namespace local = \"\"\n\n");
  
  // initial vars
  fprintf(out, "start =\n");
  assert(b->initial != NULL);
  g_hash_table_foreach(b->initial, (GHFunc) relax_ng_print_var, &t);
  fprintf(out, "\n");
  
  // productions
  assert(b->rules != NULL);
  for (i = 1; i <= b->nvars; i++)
    {
      assert(b->rules[i] != NULL);
      print_var(b, i, out);
      fprintf(out, " =\n");
      t.thrd = (void *)1;
      g_hash_table_foreach(b->rules[i], (GHFunc) relax_ng_print_rhs, &t);
      fprintf(out, "\n");
    }
}
