/* complement.c
 *
 * Implementation of the DTWA complementation alogrithm by Anca Muscholl,
 * Mathias Samuelides and Luc Segoufin.
 *
 * Copyright (C) 2009 ENS Cachan
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA.
 */

#include <limits.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

#include "bitset.h"
#include "complement.h"
#include "dtwa.h"
#include "list.h"

typedef enum {
  U, // going up in the DFS traversal
  D  // going down in the DFS traversal
} dfs;

static unsigned int nstates  = 0;
static unsigned int nstates2 = 0;
static unsigned int nreach   = 0;

/* The already constructed states. */
static bitset build_states = NULL;

/* The states to process. */
static list   states_filo  = NULL;

/* An extended state (a tuple <p,q,delta,dfs>). */
typedef unsigned long int cstate;
#define CODE_MAX ULONG_MAX

/* Extended state manipulating functions. */
static inline cstate get_code (state, state, int, dfs);
static inline state  get_state (cstate);
static inline dfs    get_dfs (cstate);
static inline const destination* get_destination (cstate, const dtwa*);
static inline state get_destination_state (cstate);
static inline int   get_destination_threshold (cstate);
static inline bool  is_last_destination (cstate, const dtwa*);

/* Register an extended state in the data structures. */
static inline void
register_state (cstate code)
{
  if (!test_bitset(build_states, code))
    {
      // fprintf(stderr, "Adding state %d\n", code);
      set_bitset(build_states, code);
      if (code != FINAL)
        states_filo = add_list(states_filo, code);
    }
}

/* Init the data structures for the complementation. */
static void
init_complement (const dtwa *wa)
{
  int i;
  nstates   = wa->nstates;
  nstates2  = nstates * nstates;
  for (i = 0; i < nstates; i++)
    if (((int) wa->reach[i][0]) > nreach)
      nreach = (int) wa->reach[i][0];
}

/* Add the initial transitions to the complement automaton `wa'. */
static void
add_init_transitions(dtwa *wa, const dtwa* automaton, bool notrans,
                     cstate size)
{
  symbol a;
  type t;
  cstate code;
  direction d;
  assert(wa != NULL);

  code = get_code(FINAL, FINAL, 1, D);
  // fprintf(stderr, "code: %lu, size: %lu, nstates: %u, nreach: %u\n", code, size, nstates, nreach);
  assert(code < size);
  register_state(code);
  if (!notrans)
    {
      d = reverse_direction(get_destination(code, automaton)->d);

      for (a = 0; a < automaton->nsyms; a++)
        for (t = ROOT; t <= LROOT; t++)
          add_transition(wa, INIT, a, t, code, d, false);
    }
}

/* The heart of the construction: adding new states and transitions. */
static void
add_code_transitions(dtwa *wa, cstate code,
                     const dtwa* automaton,
                     bool notrans,
                     cstate size)
{
  dfs       s  = get_dfs(code);
  state     p  = get_state(code);
  state     q  = get_destination(code, automaton)->s;
  direction dd = get_destination(code, automaton)->d;
  direction dm = concrete_direction(reverse_direction(dd));
  
  type t;
  symbol a;
  assert(wa != NULL);

  for (a = 0; a < automaton->nsyms; a++)
    for (t = 0; t < NTYPES; t++)
      {
        bool transition = true;
        int thres;
        cstate next_state;
        direction d;
        //fprintf(stderr, "   ... with symbol %d and type %d\n", a, t);

        // accepting state
        if (s== U && t <= LROOT && p == FINAL
            && !find_reaching_from(automaton, p,
                                   get_destination_threshold(code)+1,
                                   t))
          {
            next_state = FINAL;
            d = STAY;
          }
        // sink state
        else if (s == D && t <= LROOT && q == INIT
                 && exists_transition_to(automaton, q, a, t, p, dm))
          {
            // do nothing!
            transition = false;
          }
        // DOWN
        else if (s == D)
          {
            // going down succeeded in the DFS
            if (exists_transition_to(automaton, q, a, t, p, dm)
                && (int)automaton->reach[q][0] > 0
                && (thres = find_reaching_from(automaton, q, 1, t)))
                
              {
                next_state = get_code(q, q, thres, D);
                //fprintf(stderr, "down success: %d(%d, %d)\n", next_state, q, 1);
                d = concrete_direction (get_destination(next_state, automaton)->d);
              }
            // going down failed in the DFS
            else
              {
                next_state = get_code(p, get_destination_state(code),
                                      get_destination_threshold(code), U);
                //fprintf(stderr, "down failure: %d (%d, %d)\n", next_state, get_destination_state(code), get_destination_threshold(code));
                d = dm;
                if (!is_compatible(d, t))
                  transition = false;
              }       
          }
        // UP
        else if (s == U)
          {
            // going up and back down in the DFS
            if ((thres = find_reaching_from(automaton, p,
                                            get_destination_threshold(code)+1,
                                            t)))
              {
                next_state = get_code(p, p, thres, D);
                //fprintf(stderr, "up and down: %d(%d, %d)\n", next_state, p, get_destination_threshold(code)+1);
                d = concrete_direction (get_destination(next_state,
                                                        automaton)->d);
              }
            // going up and further up in the DFS
            else if (exists_transition_from(automaton, p, a, t))
              {
                destination *dest;
                direction dl; 
                int threshold;
                dest = automaton->trans[a][p][t];
                d = exact_direction(dest->d, t);
                dl = reverse_direction(d);
                threshold = find_reaching(automaton, dest->s, p, dl);
                assert(threshold != 0);
                next_state = get_code(dest->s, dest->s, threshold, U);
                //fprintf(stderr, "up and up: %d (%d, %d)(q%d, s%d, %d, q%d, %d--%d)\n", next_state, dest->s, threshold, p, a, t, dest->s, dest->d, dl);
                d = concrete_direction(reverse_direction(dl));
                if (!is_compatible(d, t))
                  transition = false;                  
              }
            else
              transition = false;
          }
        else
          transition = false;
        
        // add the transition
        if (transition)
          {
            assert(next_state < size);
            register_state(next_state);
            if (!notrans)
              add_transition(wa, code, a, t, next_state, d, false);
          }
      }
  
  
}

/* Build the complement of the argument DTWA. */
dtwa *
complement_dtwa(const dtwa *wa, bool names, bool notrans)
{
  dtwa *comp = NULL;
  if (wa != NULL)
    {
      cstate code;
      cstate size;
      char buffer[1024];

      init_complement(wa);
      comp = new_dtwa(false);

      // compute the maximal size
      size = nstates2;
      if (CODE_MAX / nreach > size)
        size *= nreach;
      else
        {
          fprintf(stderr, "Integer overflow: %lu * %u > %lu.\n",
                  size, nreach, CODE_MAX);
          exit(1);
        }
      if (((CODE_MAX >> 1) - 2) > size)
        size = (size << 1) + 2;
      else
        {
          fprintf(stderr, "Integer overflow: %lu * 2 + 2 > %lu.\n",
                  size, CODE_MAX);
          exit(1);
        }
      
      build_states = new_bitset (size);
      // fprintf(stderr, "Bitset size: %d\n", size);
      set_bitset(build_states, INIT);      
      add_init_transitions(comp, wa, notrans, size);

      // generate understandable names
      if (names) 
        {
          comp->states = calloc(size, sizeof(*comp->states));
          comp->states[0] = "q_f";
          comp->states[1] = "q_i";
        }

      while (states_filo != NULL)
        {
          // get the next code
          list tmp;
          code = states_filo->val;
          tmp  = states_filo;
          states_filo = states_filo->next;
          free (tmp);

          if (names)
            {
              sprintf(buffer, "[%s (%s %s) %s]",
                      get_state_name(get_state(code), wa->states),
                      get_state_name(get_destination(code, wa)->s, wa->states),
                      get_direction_name(get_destination(code, wa)->d),
                      (get_dfs(code) == U)? "^": "v");
              comp->states[code] = strdup(buffer);
            }

          //fprintf(stderr, "\n### Looking at %d [%d <%d, %d> %d]\n", code,
          //        get_state(code),
          //        get_destination(code, wa)->s,
          //        get_destination(code, wa)->d,
          //        get_dfs(code));
          add_code_transitions(comp, code, wa, notrans, size);
        }
      comp->nstates = count_bitset(build_states, size);
      comp->symbols = wa->symbols;

      free (build_states);
    }
  return comp;
}


/* Code manipulating functions. */
static inline cstate
get_code(state p, state q, int thres, dfs s)
{
  return 2 + p + nstates * (q + nstates * (thres - 1 + nreach * s));
}

static inline state
get_state(cstate code)
{
  return ((code - 2) % nstates);
}

static inline state
get_destination_state(cstate code)
{
  return ((code - 2) / nstates) % nstates;
}

static inline int
get_destination_threshold(cstate code)
{
  return (((code - 2) / nstates2) % nreach) + 1;
}

static inline dfs
get_dfs(cstate code)
{
  return (((code - 2) / nstates2) / nreach);
}

static inline const destination*
get_destination(cstate code, const dtwa* automaton)
{
  state q   = get_destination_state(code);
  int thres = get_destination_threshold(code);
  return automaton->reach[q][thres];
}

static inline bool
is_last_destination(cstate code, const dtwa* automaton)
{
  state q   = get_destination_state(code);
  int thres = get_destination_threshold(code);
  return ((int) automaton->reach[q][0]) == thres;
}
