/* dtwa.c
 *
 * Data structures and their manipulation for Deterministic Tree Walking
 * Automata.
 *
 * Copyright (C) 2009 ENS Cachan
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA.
 */

#include <assert.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libxml/hash.h>

#include "dtwa.h"

#define CHECK_ALLOC(P)\
  if ((P) == NULL)\
        {\
          fprintf(stderr, "Memory space exceeded.\n");\
          exit(1);\
        }


/* ooOo... bad! bad! bad! */
static char buffer[1024];

inline direction
reverse_direction (direction d)
{
  return (d == STAY)? STAY: (d + 3) % NDIRS;
}

inline
direction concrete_direction (direction d)
{
  return (d <= UP1)? UP: d;
}

inline direction
exact_direction (direction d, type t)
{
  if (d == UP)
    switch (t)
      {
      case L0: case C0: return UP0;
      case L1: case C1: return UP1;
      default: assert(0);
      }
    
  return d;
}

/* Checks whether a direction `d' can start from a node of type `t'. */
inline bool
is_compatible (direction d, type t)
{
  return d == STAY
    || (   (d > UP1    || t > LROOT)
        && (d < DOWN0  || t < L0)
        && (d != UP0 || ((t - 2) % 2) == 0)
        && (d != UP1 || ((t - 2) % 2) == 1));    
}

inline
bool equal_destinations (const destination *dest, state q, direction d)
{
  return dest->s == q && dest->d == d;
}

/* We store the number of transitions from a given symbol in the first
   cell.  Thanks to the usage that FINAL is 0, this does not cause any
   clash. */
static inline transition
new_tstate (transition t, state q)
{
  int max;
  assert (t != NULL);
  max = (int) t[0];   // the number of transitions in this set

  if (q > max)
    {
      int i;
      t = realloc(t, (q + 1) * sizeof(*t));
      CHECK_ALLOC(t)
      t[0] = (void *)q;
      for (i = max + 1; i <= q; i++)
        t[i] = NULL;
    }

  if (t[q] == NULL)
    {
      t[q] = calloc(NTYPES, sizeof(**t));
      CHECK_ALLOC(t[q])
    }
  
  assert (((int)t[0]) >= q);
  assert(t[q] != NULL);

  return t;
}

static inline void
new_trans (dtwa *wa, symbol s)
{
  assert(wa != NULL);

  if (s >= wa->nsyms)
    {
      int i;
      wa->trans = realloc(wa->trans, (s + 1)*sizeof(**(wa->trans)));
      CHECK_ALLOC(wa->trans)
      for (i = wa->nsyms; i <= s; i++)
        {
          wa->trans[i] = malloc(sizeof(**(wa->trans)));
          CHECK_ALLOC(wa->trans[i])
          wa->trans[i][0] = 0;
        }
      wa->nsyms = s + 1;
    }
  assert(wa->trans[wa->nsyms - 1] != NULL);
}

static inline void
new_states(dtwa *wa, state q, bool r)
{
  assert (wa != NULL);

  if (wa->nstates <= q) 
    {
      if (r)
        {
          int i;
          wa->reach = realloc (wa->reach, (q + 1) * sizeof(*wa->reach));
          CHECK_ALLOC(wa->reach)
            for (i = wa->nstates; i <= q; i++)
              {
                wa->reach[i] = malloc(sizeof(**wa->reach));
                CHECK_ALLOC(wa->reach[i])
                  wa->reach[i][0] = (destination *)0;
              }
        }
      wa->nstates = q + 1;
    }
}

static inline void
add_reachable(dtwa *wa, state q, destination *dest)
{
  int count;
  assert(wa != NULL);
  assert(wa->reach != NULL);
  assert(wa->reach[q] != NULL);

  count = (int) wa->reach[q][0];
  assert(count == 0 || wa->reach[q][count] != NULL);

  if (find_reaching(wa, q, dest->s, dest->d))
    free (dest);
  else
    {
      count++;

      // fprintf(stderr, "<%2d, ", dest->s);
      // xml_print_direction (dest->d, stderr);
      // fprintf(stderr, "> reaches %2d (at index %d).\n", q, count);
      wa->reach[q] = realloc(wa->reach[q], (count+1) * sizeof(dest));
      CHECK_ALLOC(wa->reach[q])
      wa->reach[q][0] = (destination *)count;
      wa->reach[q][count] = dest;
    }

  assert(wa->reach[q][count] != NULL);
}

/* Create a new empty DTWA.
   It contains at least two states, the initial and the final one. */
dtwa *
new_dtwa (bool r)
{
  dtwa *wa     = malloc (sizeof(*wa));
  CHECK_ALLOC(wa)

  wa->nstates = 0;
  wa->nsyms   = 0;
  wa->ntrans  = 0;
  wa->trans   = malloc (sizeof(*(wa->trans)));
  CHECK_ALLOC(wa->trans)
  if (r)
    {
      wa->reach   = malloc (sizeof(*(wa->reach)));
      CHECK_ALLOC(wa->reach)
    }
  else
    wa->reach = NULL;
  wa->symbols = NULL;
  wa->states  = NULL;
  new_states(wa, 1, r);
  
  return wa;
}

/* Add a transition to a DTWA.
   Its state, symbol, and transition sets are updated accordingly. */
bool
add_transition(dtwa *wa, state p, symbol a, type t, state q, direction d,
               bool r)
{
  destination *dest, *reach;
  bool ret = true;
  assert(wa != NULL);
  assert (p > 0);

  // check the sanity of the type vs. direction choices
  ret = ret && !(t <= LROOT && d <= UP1);
  ret = ret && !(t >= L0    && d >= DOWN0);

  if (!ret) 
    {
      fprintf(stderr,
              "warning: a transition going %s from a %s node is impossible!\n",
              get_direction_name(d), get_type_name(t));
      return ret;
    }
  
  // make sure we have allocated memory for transitions on `a'
  new_trans(wa, a);

  // make sure we have allocated memory for transitions on `a' and `p'
  wa->trans[a] = new_tstate(wa->trans[a], p);

  // test for determinism
  ret = ret && wa->trans[a][p][t] == NULL;
  if (!ret)
    fprintf(stderr, "warning: non deterministic automaton (q%d, s%d, %s)!\n",
            p, a, get_type_name(t));
  else
    {
      // build the destination
      dest = malloc(sizeof(*dest));
      CHECK_ALLOC(dest)
      dest->s = q;
      dest->d = d;
      
      // put the transition in the table
      wa->trans[a][p][t] = dest;
      
      // update the state and transition counts
      wa->ntrans++;
      if (p > q)
        new_states(wa, p, r);
      else
        new_states(wa, q, r);

      if (r)
        {
          // build the reaching pair
          reach = malloc(sizeof(*reach));
          CHECK_ALLOC(reach)
          reach->s = p;
          reach->d = reverse_direction(exact_direction(d, t));
          
          // add the reaching pair
          add_reachable(wa, q, reach);
        }

      assert(wa->trans[a][p][t] != NULL);
      return true;
    }

  return false;
}

/* Tests whether a transition exists in the DTWA. */
bool
exists_transition_from (const dtwa *wa, state p, symbol a, type t)
{
  assert (wa != NULL);
  assert (wa->trans != NULL);
  assert (wa->nsyms <= a || wa->trans[a] != NULL);

  return p > 0 && wa->nsyms > a
    && ((int)wa->trans[a][0]) >= p
    && wa->trans[a][p] != NULL
    && wa->trans[a][p][t] != NULL;
}
inline bool
exists_transition_to(const dtwa *wa, state p, symbol a, type t,
                     state q, direction d)
{
  return exists_transition_from(wa, p, a, t)
    && equal_destinations(wa->trans[a][p][t], q, d);
}

/* Lookup for a transition of form (q, d) to state p and return its
   threshold. */
int
find_reaching(const dtwa* wa, state p, state q, direction d)
{
  int i;
  int found = 0;

  assert(wa != NULL);
  assert(wa->reach != NULL);
  assert(wa->reach[p] != NULL);

  // linear search: let's hope the list isn't too long...
  for (i = 1; !found && i <= (int)wa->reach[p][0]; i++)
    if (equal_destinations(wa->reach[p][i], q, d))
      found = i;
  
  return found;
}

/* Lookup for a transition to state p starting from a threshold and
   compatible with a type, and return its threshold or 0 if not found. */ 
int
find_reaching_from(const dtwa* wa, state p, int thr, type t)
{
  int i;
  int found = 0;

  assert(wa != NULL);
  assert(wa->reach != NULL);
  assert(wa->reach[p] != NULL);

  for (i = thr; !found && i <= (int)wa->reach[p][0]; i++)
    if (is_compatible(wa->reach[p][i]->d, t))
      found = i;
  
  return found;
}

/* Destroy a DTWA, freeing unused space. */
void
free_dtwa (dtwa *wa)
{
  int i, j, k;
  assert(wa != NULL);

  for (i = 0; i < wa->nsyms; i++)
    {
      for (j = 1; j <= (int) wa->trans[i][0]; j++)
        if (wa->trans[i][j] != NULL)
          {
            for (k = 0; k < NTYPES; k++)
              if (wa->trans[i][j][k] != NULL)
                free(wa->trans[i][j][k]);
            free(wa->trans[i][j]);
          }
      free(wa->trans[i]);
    }
  free(wa->trans);

  if (wa->reach != NULL)
    {
      for (i = 0; i < wa->nstates; i++)
        {
          for (j = 1; j <= (int) wa->reach[i][0]; j++)
            free (wa->reach[i][j]);
          free (wa->reach[i]);
        }
      free (wa->reach);
    }

  if (wa->symbols != NULL)
    xmlHashFree ((xmlHashTablePtr) wa->symbols,  NULL);    

  if (wa->states != NULL)
    free(wa->states);

  free(wa);
  wa = NULL;      
}

inline char *
get_state_name(state s, char **states)
{
  if (states == NULL)
    {
      sprintf(buffer, "q%u", s);
      return buffer;
    }
  return states[s];
}

static inline void
xml_print_state(state s, FILE *out, char **states)
{
  fprintf(out, "<state name=\"%s\"/>", get_state_name(s, states));
}

static inline void
xml_print_symbol(symbol s, FILE *out, char **syms)
{
  if (syms == NULL)
    fprintf(out, "<symbol name=\"s%u\"/>", s);
  else
    fprintf(out, "<symbol name=\"%s\"/>", syms[s]);
}

inline char *
get_type_name(type t)
{
  switch (t)
    {
    case ROOT:  return "root";
    case LROOT: return "lroot";
    case L0:    return "leaf0";
    case L1:    return "leaf1";
    case C0:    return "child0";
    case C1:    return "child1";
    default:    assert(0);
    }
}

inline char *
get_direction_name(direction d)
{
  switch (d)
    {
    case UP:    return "up";
    case STAY:  return "stay";
    case DOWN0: return "down0";
    case DOWN1: return "down1";
    case UP0:   return "up0";
    case UP1:   return "up1";
    default:    assert(0);
    }
}

static void
sym_copy (void * payload, void * data, xmlChar * name)
{
  char **syms = (char**) data;
  syms[(int)payload - 1] = (char*) name;
}


/* Prints out a DTWA as an XML string. */
void
xml_print_dtwa(const dtwa *wa, FILE *out)
{  
  int i, j, k;
  char **syms = NULL;
  assert(wa != NULL);

  // preprocess symbols
  if (wa->symbols != NULL)
    {
      xmlHashTablePtr hash = (xmlHashTablePtr) wa->symbols;
      syms = calloc(xmlHashSize(hash), sizeof(*syms));
      CHECK_ALLOC(syms)
      xmlHashScan(hash, sym_copy, syms);
    }

  fprintf(out, "<?xml version=\"1.0\"?>\n");
  fprintf(out, "<automaton states=\"%d\" symbols=\"%d\" transitions=\"%d\">\n",
          wa->nstates, wa->nsyms, wa->ntrans);

  fprintf(out, "  <init>");
  xml_print_state(INIT, out, wa->states);
  fprintf(out, "</init>\n");
  fprintf(out, "  <final>");
  xml_print_state(FINAL, out, wa->states);
  fprintf(out, "</final>\n");  

  fprintf(out, "  <transitions>\n");
  for (i = 0; i < wa->nsyms; i++)
    for (j = 1; j <= (int) wa->trans[i][0]; j++)
      if (wa->trans[i][j] != NULL)
        for (k = 0; k < NTYPES; k++)
          if (wa->trans[i][j][k] != NULL)
            {
              fprintf(out, "    <trans type=\"%s\"", get_type_name(k));
              fprintf(out, " direction=\"%s\">\n",
                      get_direction_name(wa->trans[i][j][k]->d));
              
              fprintf(out, "      <from>");
              xml_print_state(j, out, wa->states);
              fprintf(out, "</from>\n");
              
              fprintf(out, "      ");
              xml_print_symbol(i, out, syms);
              fprintf(out, "\n");
              
              fprintf(out, "      <to>");
              xml_print_state((wa->trans[i][j][k])->s, out, wa->states);
              fprintf(out, "</to>\n");
              
              fprintf(out, "    </trans>\n");
            }
  fprintf(out, "  </transitions>\n");
  fprintf(out, "</automaton>\n");
  
  if (wa->symbols != NULL)   
    free(syms);
}
