/* dtwa.h
 *
 * Data structures and their manipulation for Deterministic Tree Walking
 * Automata.
 *
 * Copyright (C) 2009 ENS Cachan
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA.
 */

#ifndef DTWA_H
#define DTWA_H

#include <assert.h>
#include <stdbool.h>
#include <stdio.h>

/*------------------------------------------- Type of tree nodes. */
typedef enum {
  ROOT,      // 0: the tree root
  LROOT,     // 1: a leaf root, i.e. the root of tree with a single node
  C0,        // 2: a left interior node
  C1,        // 3: a right interior node
  L0,        // 4: a left leaf
  L1         // 5: a right leaf
} type;
#define NTYPES 6

/* Return a string representation of this type. */
inline char *get_type_name (type);

/*------------------------------------------ Directions in trees. */
typedef enum {
  UP,        // 0: just go up...
  UP0,       // 1: go up from a left child
  UP1,       // 2: go up from a right child
  STAY,      // 3: stay in place
  DOWN0,     // 4: go down to the left child
  DOWN1      // 5: go down to the right child
} direction;
#define NDIRS 6

/* Return a string representation of this type. */
inline char *get_direction_name (direction);

/* Return the opposite direction. */
inline direction reverse_direction (direction);

/* Return the concrete direction. */
inline direction concrete_direction (direction);

/* Return the exact direction. */
inline direction exact_direction (direction, type);

/* Tell whether a transition with direction `d' can start from a node of
   type `t'. */
inline bool is_compatible (direction d, type t);

/*------------------------------------------ Definition of DTWAs. */
/* A finite alphabet: just numbers, dude. */
typedef unsigned int symbol;

/* A finite set of states: guess what that is... */
typedef unsigned int state;
#define INIT  1
#define FINAL 0

/* Get a string representation for this state. */
inline char *get_state_name(state, char **);

/* The destination of a transition is a pair (state, direction). */
typedef struct
{
  state     s;
  direction d;
} destination;

inline bool equal_destinations (const destination*, state, direction);


/* A transition set from a symbol takes you from a pair (state, type) to a
   destination. */
typedef destination ***transition;

/* A Deterministic Tree Walking Automaton

   We consider normalized DTWAs where there is a unique final state with
   no outgoing transitions.  By convention, the initial state is `INIT'
   and the final one `FINAL'.  We reserve `0' for the final state,
   and *no transition* should be made from that state.

   The destination of a transition of a DTWA `wa' from state `p' on
   symbol `a' on a node of type `t' is `wa->trans[a][p][t]'.

   The pairs (p, d) that reach a state q are in the `wa->reach[q]'
   table, with `wa->reach[q][(int)wa->reach[q][0]]' elements starting
   from index `1'.
*/
typedef struct
{
  int nstates;          // the states are numbered from `0' to `nstates - 1'
  int nsyms;            // the symbols are numbered from `0' to `nsyms - 1'
  int ntrans;           // the number of transitions
  transition* trans;    // the transition table
  destination*** reach; // which pairs (state, direction) reach each state
  void *symbols;        // the names of the symbols
  char **states;        // the names of the states
} dtwa;



/*------------------------------------------ Operations on DTWAs. */
/* Create a new empty DTWA.
   It contains at least two states, the initial and the final one. */
dtwa *new_dtwa(bool);

/* Add a transition to a DTWA, returning 0 or 1 for failure or success.
   Its state, symbol, and transition sets are updated accordingly. */
bool add_transition(dtwa*, state, symbol, type, state, direction, bool);

/* Tests whether a transition exists in the DTWA. */
bool exists_transition_from(const dtwa *, state, symbol, type);
inline bool exists_transition_to(const dtwa*, state, symbol, type,
                                 state, direction);

/* Lookup for a transition of form (q, d) to state p and return its
   threshold. */
int find_reaching(const dtwa*, state, state, direction);

/* Lookup for a transition to state p starting from a threshold and
   compatible with a type, and return its threshold or 0 if not found. */ 
int find_reaching_from(const dtwa*, state, int, type);

/* Destroy a DTWA, freeing unused space. */
void free_dtwa (dtwa *);

/* Prints out a DTWA as an XML string. */
void xml_print_dtwa(const dtwa *, FILE *);

#endif /* DTWA_H */
