/* dtwa2rng.c
 *
 * Transform DTWAs into equivalent RELAX NG grammars.
 *
 * Copyright (C) 2009 ENS Cachan
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "branching.h"
#include "buta.h"
#include "dtwa.h"
#include "parse.h"

static void
version (FILE *out)
{
  fprintf(out, "dtwa2rng version %s - Transform DTWAs into equivalent RELAX NG grammars.\n", PACKAGE_VERSION);
  fprintf(out, "\n");
  fprintf(out, "Copyright (C) 2009 ENS Cachan\n");
  fprintf(out, "This is free software; see the source for copying conditions.  There is NO\n");
  fprintf(out, "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
}

static void
usage (FILE *out)
{
  fprintf(out, "Usage:\n");
  fprintf(out, "  dtwa2rng [OPTIONS] - transform DTWAs into equivalent RELAX NG grammars\n");
  fprintf(out, "\n");
  fprintf(out, "  Computes an equivalent RELAX NG grammar for the automaton provided on the standard input and output the result on standard output.\n");
  fprintf(out, "\n");
  fprintf(out, "Help Options:\n");
  fprintf(out, "  -?, --help           Show help options\n");
  fprintf(out, "\n");
  fprintf(out, "Application Options:\n");
  fprintf(out, "  -n, --names          Give explicit names to states (uses a lot of memory)\n");
  fprintf(out, "  -t, --no-transitions Do not generate transitions in the resulting automaton\n                       (only return the number of variables)\n");
  fprintf(out, "  -V, --version        Output version information and exit\n");
}

int
main (int argc, char *argv[])
{
  int i;
  buta *rng;
  dtwa *wa;
  bool names   = false;
  bool notrans = false;

  // process options
  for (i = 1; i < argc; i++)
    {
      // process options
      if (strcmp(argv[i], "-n") == 0 || strcmp(argv[i], "--names") == 0)
        names = true;
      else if (strcmp(argv[i], "-t") == 0
          || strcmp(argv[i], "--no-transitions") == 0)
        notrans = true;
      else if (strcmp(argv[i], "-?") == 0
               || strcmp(argv[i], "-h") == 0
               || strcmp(argv[i], "--help") == 0)
        {
          usage(stdout);
          return 0;
        }
      else if (strcmp(argv[i], "-V") == 0 || strcmp(argv[i], "--version") == 0)
        {
          version(stdout);
          return 0;
        }
      else
        {
          fprintf(stderr, "dtwac: unknown option `%s'.\n", argv[i]);
          usage(stderr);
        }
    }

  wa   = parse(0, "stdin", DTWA_DTD);
  rng = dtwa2buta (wa, names, notrans);
  //rng = new_buta();
  //add_rule(rng, 0, 0, 0, 1);
  //add_rule(rng, 0, 1, 1, 1);
  //add_rule(rng, 1, 1, 1, 1);
  //set_initial(rng, 1);

  if (notrans)
    printf("%u\n", rng->nvars);
  else
    relax_ng_print_buta (rng, stdout);
  
  free_dtwa (wa);
  free_buta (rng);

  return 0;  
}
