/* dtwarun.c
 *
 * Main entry point for running DTWAs.
 *
 * Copyright (C) 2009 ENS Cachan
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>

#include <libxml/tree.h>

#include "dtwa.h"
#include "parse.h"
#include "run.h"

static void
version (FILE *out)
{
  fprintf(out, "dtwarun version %s - An interpreter for deterministic tree walking automata.\n", PACKAGE_VERSION);
  fprintf(out, "\n");
  fprintf(out, "Copyright (C) 2009 ENS Cachan\n");
  fprintf(out, "This is free software; see the source for copying conditions.  There is NO\n");
  fprintf(out, "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
}

static void
usage (FILE *out)
{
  fprintf(out, "Usage:\n");
  fprintf(out, "  dtwarun [OPTIONS] [FILE] - deterministic tree walking automaton interpreter\n");
  fprintf(out, "\n");
  fprintf(out, "  Runs the automaton provided through the `--automaton' option on the XML tree\n  of FILE, or on the standard input if no FILE is provided.\n");
  fprintf(out, "\n");
  fprintf(out, "Help Options:\n");
  fprintf(out, "  -?, --help          Show help options\n");
  fprintf(out, "\n");
  fprintf(out, "Application Options:\n");
  fprintf(out, "  --automaton FILE    Read the DTWA automaton from FILE\n");
  fprintf(out, "  -v, --verbose       Run the automaton in verbose mode\n");
  fprintf(out, "  -V, --version       Output version information and exit\n");
}

int
main (int argc, char *argv[])
{
  int i = 1;
  char *autofile = NULL;
  char *xmlfile  = NULL;
  bool  verbose  = false;

  if (argc < 2)
    {
      fprintf(stderr, "%s: missing argument.\n", argv[0]);
      usage(stderr);
      exit(1);
    }
  // process options
  for (i = 1; i < argc ; i++)
    {
      if (argv[i][0] != '-')
 	xmlfile = argv[i];
      else if (strcmp(argv[i], "-?") == 0 || strcmp(argv[i], "--help") == 0)
        {
          usage(stdout);
          return 0;
        }
      else if (strcmp(argv[i], "-v") == 0 || strcmp(argv[i], "--verbose") == 0)
        verbose = true;
      else if (strcmp(argv[i], "-V") == 0 || strcmp(argv[i], "--version") == 0)
        {
          version(stdout);
          return 0;
        }
      else if (strcmp(argv[i], "--automaton") == 0)
        autofile = argv[++i];
      else if (strlen(argv[i]) > 12 && argv[i][11] == '='
               && !(argv[i][11] = 0)
               && (strcmp(argv[i], "--automaton") == 0 || (argv[i][11] = '=')))
        autofile = argv[i] + 12;
      else
        fprintf(stderr, "%s: unknown option `%s'.\n", argv[0], argv[i]);
    }

  // process files
  if (autofile == NULL)
    {
      fprintf(stderr, "%s: missing `--automaton' option.\n", argv[0]);
      usage(stderr);
      exit(1);
    }
  else
    {
      int infd;
      dtwa *wa;
      xmlDocPtr doc;
      bool accept = false;

      // parse automaton XML
      infd = open(autofile, O_RDONLY);
      if (infd == -1)
        {
          fprintf(stderr, "%s: unable to read from `%s'.\n", argv[0],
                  autofile);
          exit(1);
        }
      
      wa  = parse(infd, autofile, DTWA_DTD);
      close(infd);

      // parse input XML
      if (!xmlfile)
        {
          infd = 0;
          xmlfile = "stdin";
        }
      else
        {
          infd = open(xmlfile, O_RDONLY);
          if (infd == -1)
            {
              fprintf(stderr, "%s: unable to read from `%s'.\n", argv[0],
                      xmlfile);
              exit(1);
            }
        }
      doc = xml_tree(infd, xmlfile);
      if (doc == NULL)
        {
          fprintf(stderr, "%s: unable to parse `%s'.\n", argv[0],
                  xmlfile);
          exit(1);
        }
      if (infd)
        close(infd);

      // run the automaton
      if (verbose)
        printf("Running `%s' on `%s'.\n\n", autofile, xmlfile);
      accept = run(wa, doc, &std_nav, verbose);
      if (verbose)
        printf("\nResult of run: %s.\n", (accept)? "success": "failure");

      // free memory
      xmlFreeDoc(doc);
      free_dtwa (wa);

      return accept;
    }

  return 1;  
}
