/* parse.c
 *
 * XML parsing using libxml2.  The validation portion of the code is
 * inspired by `xmllint.c' from the libxml2 distribution, but does not
 * really fall under the range of a "substantial" portion of code...
 *
 * Copyright (C) 2009 ENS Cachan
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA.
 */

#include <assert.h>
#include <stdbool.h>
#include <string.h>
#include <libxml/hash.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/valid.h>
#include <libxml/xmlerror.h>
#include <libxml/xmlstring.h>
#include <libxml/xpath.h>

#include "dtwa.h"
#include "parse.h"

#define HASH_SIZE 127

/* XML utilities. */
static xmlNodePtr
attrNodeWithName (const xmlNodePtr node, const xmlChar *name)
{
  xmlAttrPtr attr;
  assert (node != NULL);

  for (attr = node->properties; attr; attr = attr->next)
    if (xmlStrEqual(attr->name, name))
      return attr->children;

  return NULL;
}

inline xmlChar *
attributeWithName (const xmlNodePtr node, const xmlChar *name)
{
  xmlNodePtr attr = attrNodeWithName(node, name);
  return (attr != NULL)? attr->content: NULL;
}

xmlNodePtr
siblingWithName (const xmlNodePtr node, const xmlChar *name)
{
  xmlNodePtr cur;

  for (cur = node; cur; cur = cur->next)
    if (cur->type == XML_ELEMENT_NODE && xmlStrEqual(cur->name, name))
      return cur;
  
  return NULL;
}

inline xmlNodePtr
childWithName (const xmlNodePtr node, const xmlChar *name)
{
  assert (node != NULL);
  return siblingWithName(node->children, name);
}

unsigned int
countSiblings (const xmlNodePtr node)
{
  unsigned int ret = 0;
  xmlNodePtr cur;
  assert (node != NULL);

  for (cur = node; cur; cur = cur->next)
    if (cur->type == XML_ELEMENT_NODE)
      ret++;
  
  return ret;
}

inline unsigned int
countChildren (const xmlNodePtr node)
{
  assert(node != NULL);
  return countSiblings(node->children);
}

/* Validate a document against the DTD. */
static bool
validate_dtwa (const xmlDocPtr doc, const char* filename, const char* dtdname)
{
  bool valid = true;
  xmlDtdPtr dtd;

  if (doc == NULL)
    return false;

  dtd = xmlParseDTD(NULL, (const xmlChar*)dtdname);
  if (dtd == NULL)
    xmlGenericError(xmlGenericErrorContext, "Could not parse DTD %s.\n",
                    dtdname);
  else
    {
      xmlValidCtxtPtr cvp = xmlNewValidCtxt();
      
      if (cvp == NULL)
        {
          xmlGenericError(xmlGenericErrorContext,
                          "Couldn't allocate validation context.\n");
          exit(-1);
 	}
      cvp->userData = (void *) stderr;
      cvp->error    = (xmlValidityErrorFunc)   fprintf;
      cvp->warning  = (xmlValidityWarningFunc) fprintf;

      valid = xmlValidateDtd(cvp, doc, dtd);
      
      if (!valid)
        xmlGenericError(xmlGenericErrorContext,
                        "Document %s does not validate against %s.\n",
                        filename, dtdname);
      xmlFreeValidCtxt(cvp);
      xmlFreeDtd(dtd);
    }
  return valid;
}

/* Hash table manipulation. */
static inline unsigned int
add_name(const xmlNodePtr s, xmlHashTablePtr hash)
{
  state p;
  xmlChar *name;
  assert(s != NULL);
  assert(hash != NULL);

  name = attributeWithName(s, BAD_CAST "name");
  assert(name != NULL);
  if (!(p = (int)xmlHashLookup(hash, name)))
    {
      p = xmlHashSize(hash) + 1;
      xmlHashAddEntry(hash, name, (void*)p);
      // fprintf(stderr, "%s becomes q%d\n", name, p-1);
    }
  return p;
}

/* Check whether the automaton is normalized or not. */
static int
check_normalized(const xmlDocPtr doc, xmlHashTablePtr hash)
{
  int ret;
  xmlXPathContextPtr xpathCtx;
  xmlXPathObjectPtr  xpathObj;
  xmlNodeSetPtr      nodes;
  const xmlChar *    xpathExpr = BAD_CAST
  "//final/state[count(//final/state) = 1 and not(//from/state/@name = //final/state/@name)]";

  assert(doc  != NULL);
  assert(hash != NULL);

  xpathCtx = xmlXPathNewContext(doc);
  if (xpathCtx == NULL)
    {
      xmlGenericError(xmlGenericErrorContext,
                      "Error: unable to create new XPath context.\n");
      exit(-1);
    }

  xpathObj = xmlXPathEvalExpression(xpathExpr, xpathCtx);
  if(xpathObj == NULL)
    {
      xmlGenericError(xmlGenericErrorContext,
                      "Error: unable to evaluate XPath expression \"%s\".\n",
                      xpathExpr);
      exit(-1);
    }

  nodes = xpathObj->nodesetval;
  ret = nodes->nodeNr;
#ifndef NDEBUG
  if (ret)
    {
      int p = add_name(nodes->nodeTab[0], hash);
      assert(p == (FINAL + ret));
    }
#endif
  xmlXPathFreeObject(xpathObj);
  xmlXPathFreeContext(xpathCtx); 

  return ret;
}

/* Building functions. */
static void
process_initial_state(const xmlNodePtr automaton,
                      xmlHashTablePtr hash,
                      int norm)
{
  xmlNodePtr cur;
  state ret;
  assert(automaton != NULL);
  assert(hash != NULL);

  cur  = childWithName(automaton, BAD_CAST "init");
  assert(cur != NULL);
  cur  = childWithName(cur, BAD_CAST "state");  
  ret = add_name(cur, hash);
  assert(ret == (INIT + norm));
}

static inline direction
process_direction (xmlChar* dir)
{
  assert(dir != NULL);
  if (xmlStrEqual(dir, BAD_CAST "up"))
      return UP;
  if (xmlStrEqual(dir, BAD_CAST "stay"))
      return STAY;
  if (xmlStrEqual(dir, BAD_CAST "down0"))
      return DOWN0;
  if (xmlStrEqual(dir, BAD_CAST "down1"))
      return DOWN1;
  assert(false);
  return 0;
}

static inline type
process_type (xmlChar* type)
{
  assert(type != NULL);
  if (xmlStrEqual(type, BAD_CAST "root"))
      return ROOT;
  if (xmlStrEqual(type, BAD_CAST "lroot"))
      return LROOT;  
  if (xmlStrEqual(type, BAD_CAST "child0"))
      return C0;
  if (xmlStrEqual(type, BAD_CAST "child1"))
      return C1;
  if (xmlStrEqual(type, BAD_CAST "leaf0"))
      return L0;
  if (xmlStrEqual(type, BAD_CAST "leaf1"))
      return L1;
  assert(false);
  return 0;
}

static void
process_transitions(dtwa *wa,
                    const xmlNodePtr automaton,
                    xmlHashTablePtr states,
                    xmlHashTablePtr symbols,
                    int norm)
{
  xmlNodePtr cur;
  assert(wa != NULL);
  assert(automaton != NULL);
  assert(states != NULL);
  assert(symbols != NULL);

  cur = childWithName(automaton, BAD_CAST "transitions");
  assert(cur != NULL);
  for (cur = childWithName(cur, BAD_CAST "trans");
       cur != NULL;
       cur = siblingWithName(cur->next, BAD_CAST "trans"))
    {
      xmlChar *typa = attributeWithName(cur, BAD_CAST "type");
      xmlChar *dir  = attributeWithName(cur, BAD_CAST "direction");
      xmlNodePtr f  = childWithName(cur, BAD_CAST "from");
      xmlNodePtr s  = childWithName(cur, BAD_CAST "symbol");
      xmlNodePtr to = childWithName(cur, BAD_CAST "to");

      state p, q;
      symbol a;
      type t;
      direction d;

      assert(typa && dir && f && s && to);
      p = add_name(childWithName(f,  BAD_CAST "state"), states) - norm;
      q = add_name(childWithName(to, BAD_CAST "state"), states) - norm;
      a = add_name(s, symbols) - 1;
      t = process_type(typa);
      d = process_direction(dir);
      //fprintf(stderr, "  transition q%d -%d-> q%d\n", p, a, q);

      add_transition(wa, p, a, t, q, d, true);
    }
}

static void
process_final_states(dtwa *wa,
                     const xmlNodePtr automaton,
                     xmlHashTablePtr states)
{
  xmlNodePtr cur;
  assert(wa != NULL);
  assert(automaton != NULL);
  assert(states != NULL);

  cur  = childWithName(automaton, BAD_CAST "final");
  assert(cur != NULL);
  for (cur = childWithName(cur, BAD_CAST "state");
       cur != NULL;
       cur = siblingWithName(cur->next, BAD_CAST "state"))
    {
      state p;
      symbol a;
      type t;
      
      p = add_name(cur, states);
      for (a = 0; a < wa->nsyms; a++)
        for (t = ROOT; t <= LROOT; t++)
          {
            if (exists_transition_from (wa, p, a, t))
              {
                free (wa->trans[a][p][t]);
                wa->trans[a][p][t] = NULL;
              }
            //fprintf(stderr, "  transition q%d -%d-> q%d\n", p, a, FINAL);
            add_transition(wa, p, a, t, FINAL, STAY, true);
          }
    }
}

static void
sym_copy (void * payload, void * data, xmlChar * name)
{
  char **syms = (char**) data;
  syms[(int)payload] = strdup((char*) name);
}

static void
sym_copy_norm (void * payload, void * data, xmlChar * name)
{
  char **syms = (char**) data;
  syms[(int)payload - 1] = strdup((char*) name);
}

/* Build a DTWA from an XML document. */
static dtwa *
xml_to_dtwa(const xmlDocPtr doc)
{
  int norm;
  dtwa *wa;
  xmlNodePtr automaton;
  xmlHashTablePtr states;
  xmlHashTablePtr symbols;
  assert(doc != NULL);

  /* Create the DTWA and the hash table. */
  wa      = new_dtwa (true);
  states  = xmlHashCreate (HASH_SIZE); // here we could look at the
                                       // "states" attribute in the
                                       // document and use it as an
                                       // initialization value...
  symbols = xmlHashCreate (HASH_SIZE); // same here with "symbols"...
  automaton = siblingWithName(doc->children, BAD_CAST "automaton");
  if (wa == NULL || states == NULL || symbols == NULL)
    {
      xmlGenericError(xmlGenericErrorContext,
                      "Couldn't allocate memory.\n");
      exit(-1);
    }
  if (automaton == NULL)
    {      
      xmlGenericError(xmlGenericErrorContext,
                 "No automaton element found at the root of the document.\n");
      exit(-1);
    }

  /* Populate the hash table. */
  norm = check_normalized(doc, states);
  // fprintf(stderr, "Processing initial state...\n");
  process_initial_state(automaton, states, norm);
  // fprintf(stderr, "Processing transitions...\n");
  process_transitions(wa, automaton, states, symbols, norm);
  if (!norm)
    {
      // fprintf(stderr, "Processing final states...\n");
      process_final_states (wa, automaton, states);
    }

  wa->symbols = symbols;
  wa->states  = calloc(xmlHashSize(states)+1, sizeof(*wa->states));
  if (!norm)
    {
      xmlHashScan(states, sym_copy, wa->states);
      wa->states[0] = "__final";
    }
  else
    xmlHashScan(states, sym_copy_norm, wa->states);
    

  /* Free the hash table. */
  xmlHashFree (states,  NULL);
  
  return wa;
}

/* The main parsing function. */
dtwa *
parse(int infd, const char *infile, const char* dtdname)
{
  xmlDocPtr dtwadoc = xml_tree(infd, infile);
  dtwa *wa;

  if (!validate_dtwa(dtwadoc, infile, dtdname))
    {
      xmlGenericError(xmlGenericErrorContext,
                      "dtwac: unable to parse file `%s'.\n",
                      infile);
      exit(1);
    }

  /* Process XML document. */
  wa = xml_to_dtwa(dtwadoc);

  /* Freeing... */
  xmlFreeDoc(dtwadoc);

  return wa;
}

xmlDocPtr
xml_tree(int infd, const char *infile)
{
  xmlDocPtr doc;

  LIBXML_TEST_VERSION;
  xmlInitParser();

#if LIBXML_VERSION >= 20621
  doc = xmlReadFd(infd, infile, NULL, XML_PARSE_COMPACT);
#else
  doc = xmlReadFd(infd, infile, NULL, 0);
#endif
  xmlCleanupParser();

  return doc;
}
