/* run.h
 *
 * Run a DTWA on an XML tree.
 *
 * Copyright (C) 2009 ENS Cachan
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301  USA.
 */

#include <stdbool.h>
#include <libxml/hash.h>
#include <libxml/tree.h>

#include "dtwa.h"
#include "parse.h"
#include "run.h"

/* Run the automaton against the document. */
bool
run(dtwa* wa, const xmlDocPtr doc, tree_nav* nav, bool verbose)
{
  xmlNodePtr node;
  xmlHashTablePtr symbols;
  state      p;
  bool       accept = false;
  bool       ended  = false;
  assert(wa != NULL);
  assert(doc != NULL);
  assert(nav != NULL);
  assert(wa->symbols != NULL);

  symbols = (xmlHashTablePtr) wa->symbols;

  // get the root node
  for (node = doc->children; node; node = node->next)
    if (node->type == XML_ELEMENT_NODE)
      break;
  
  if (node == NULL)
    {
      fprintf(stderr, "Empty document.\n");
      exit(1);
    }
  assert(nav->typ(node) <= LROOT);

  // run the automaton
  p = INIT;
  while (node != NULL && !ended)
    {
      type t = nav->typ(node);
      
      if (p == FINAL && t <= LROOT)
        {
          accept = true;
          ended  = true;
        }
      else
        {
          symbol a = ((symbol)xmlHashLookup(symbols, node->name)) - 1;
          assert(a >= 0);
          if (!exists_transition_from(wa, p, a, t))
            {
              accept = false;
              ended  = true;
            }
          else
            {
              destination *dest = wa->trans[a][p][t];
              if (verbose)
                printf("%s \t--%s, %s",
                       get_state_name(p, wa->states),
                       node->name,
                       get_type_name(t));
              p    = dest->s;
              node = (nav->nav[dest->d])(node);
              if (verbose)
                printf(", %s-->\t%s",
                       get_direction_name(dest->d),
                       get_state_name(p, wa->states));
            }
        }
      if (verbose)
        printf("\n");
    }
  return accept;
}

/* Navigation functions. */
static xmlNodePtr
stay (const xmlNodePtr n)
{
  return n;
}

/* Standard navigation functions. */
static xmlNodePtr
std_up (const xmlNodePtr n)
{
  if (n == NULL)
    return NULL;
  return n->parent;
}

static xmlNodePtr
std_down0 (const xmlNodePtr n)
{
  if (n != NULL && n->children != NULL)
    {
      xmlNodePtr cur;
      for (cur = n->children; cur; cur = cur->next)
        if (cur->type == XML_ELEMENT_NODE)
          return cur;      
    }
  return NULL;
}

static xmlNodePtr
std_down1 (const xmlNodePtr n)
{
  xmlNodePtr cur = std_down0(n);
  if (cur != NULL)
    for (cur = cur->next; cur; cur = cur->next)
      if (cur->type == XML_ELEMENT_NODE)
        break;
  return cur;
}

static type
std_type (const xmlNodePtr n)
{
  bool interior, root, first;
  xmlNodePtr cur;
  assert (n != NULL);

  interior  = (n->children != NULL && countChildren(n) > 0);
  root      = (n == n->doc->children);
  for (cur = n->prev; cur; cur = cur->prev)
      if (cur->type == XML_ELEMENT_NODE)
        break;
  first     = (cur == NULL);

  if (root && interior)
    return ROOT;
  if (root)
    return LROOT;
  if (first && interior)
    return C0;
  if (interior)
    return C1;
  if (first)
    return L0;
  return L1;
}

tree_nav std_nav = { { std_up, NULL, NULL, stay, std_down0, std_down1 },
                     std_type };
